@@ -1,5 +1,5 @@
 /*
- * powernowd.c: (c) 2003-2004 John Clemens <clemej@alum.rpi.edu>
+ * powernowd.c: (c) 2003-2005 John Clemens <clemej@alum.rpi.edu>
  *
  * Released under the GNU General Public License. See the LICENSE file
  * included with this file.
@@ -22,12 +22,22 @@
  *   	    old pause/unpause behavior a cleaner way. Added LEAPS mode and 
  *   	    verbosity patches by Hans Ulrich Niedermann. Cleaned up verbosity
  *   	    handling more.
- *   	    
+ *   v0.95, Added proper HT support, and SMT/CMP infrastructure. Hope it 
+ *   	    works. Added scalaing_available_frequencies support and made 
+ *   	    default.  
+ *   v0.96  Fix HT detection code to default to 1 if cpuid gives strange
+ *          result (fixes centrinos). Cleaned up error reporting in 
+ *          read_file() (fixes non-SAF drivers. i.e. pmac).
+ *   
  * * Contributions from:
  * 	Warren Togami <warren@togami.com>
  * 	Michael Schreiter <Michael_Schreiter@gmx.de>
  * 	Wolfgang Tremmel <tremmel@garf.de>
  * 	Hans Ulrich Niedermann <kdev@n-dimensional.de>
+ * 	-- And many others who have sent me patches over the past 3 years,
+ * 	   some of whom I've never even acknowledged.  Just know that your 
+ * 	   support is appreciated, even if i didn't merge in your ideas.  
+ * 	   Thank you.
  */
 
 
@@ -57,6 +67,9 @@ typedef struct cpustats {
 	unsigned long long mynice;
 	unsigned long long system;
 	unsigned long long idle;
+	unsigned long long iowait;
+	unsigned long long irq;
+	unsigned long long softirq;
 } cpustats_t;
 
 typedef struct cpuinfo {
@@ -65,11 +78,15 @@ typedef struct cpuinfo {
 	unsigned int max_speed;
 	unsigned int min_speed;
 	unsigned int current_speed;
+	unsigned int speed_index;
 	int fd;
 	char *sysfs_dir;
 	cpustats_t *last_reading;
 	cpustats_t *reading;
-	int in_mhz; /* 0 = speed in kHz, 1 = speed in mHz */  
+	int in_mhz; /* 0 = speed in kHz, 1 = speed in mHz */
+	unsigned long *freq_table;
+	int table_size;
+	int threads_per_core;
 } cpuinfo_t;
 
 /* 
@@ -90,8 +107,9 @@ enum function {
 } func = AGGRESSIVE; 
 
 enum modes {
-	RAISE,
-	LOWER
+	LOWER,
+	SAME,
+	RAISE
 };
 
 /* for a daemon as simple as this, global data is ok. */
@@ -103,6 +121,11 @@ unsigned int step = 100000;  /* in kHz */
 unsigned int poll = 1000; /* in msecs */
 unsigned int highwater = 80;
 unsigned int lowwater = 20;
+unsigned int max_limit = 0;
+unsigned int min_limit = 0;
+unsigned int step_specified = 0;
+unsigned int t_per_core = 1;
+unsigned int cores_specified = 0;
 
 /* statistics */
 unsigned int change_speed_count = 0;
@@ -110,14 +133,13 @@ time_t start_time = 0;
 
 #define SYSFS_TREE "/sys/devices/system/cpu/"
 #define SYSFS_SETSPEED "scaling_setspeed"
-#define SYSFS_GOVERNOR "scaling_governor"
 
-#define VERSION	"0.90"
+#define VERSION	"0.96"
 
 void help(void)
 {
 	
-	printf("PowerNow Daemon v%s, (c) 2003-2004 John Clemens\n", VERSION);
+	printf("PowerNow Daemon v%s, (c) 2003-2005 John Clemens\n", VERSION);
 	printf("Daemon to control the speed and voltage of cpus.\n\n");
 	printf("This is a simple client to the CPUFreq driver, and uses\n");
 	printf("linux kernel v2.5 sysfs interface.  You need a supported\n");
@@ -134,6 +156,7 @@ void help(void)
 	printf("		2 = PASSIVE, 3 = LEAPS\n");
 	printf("	-s #	Frequency step in kHz (default = 100000)\n");
 	printf("	-p #	Polling frequency in msecs (default = 1000)\n");
+	printf("	-c #	Specify number of threads per power-managed core\n");
 	printf("	-u #	CPU usage upper limit percentage [0 .. 100, default 80]\n");
 	printf("	-l #    CPU usage lower limit percentage [0 .. 100, default 20]\n");
 
@@ -152,7 +175,7 @@ int read_file(const char *file, int fd, int new)
 	if (new) {
 		if ((fd = open(file, O_RDONLY)) == -1) {
 			err = errno;
-			perror("Couldn't open file");
+			perror(file);
 			return err;
 		}
 	}
@@ -160,7 +183,7 @@ int read_file(const char *file, int fd, int new)
 	lseek(fd, 0, SEEK_SET);
 	if ((n = read(fd, buf, sizeof(buf)-1)) < 0) {
 		err = errno;
-		perror("Couldn't read file");
+		perror(file);
 		close(fd);
 		return err;
 	}
@@ -177,15 +200,14 @@ int read_file(const char *file, int fd, int new)
  *
  * Format of line:
  * ...
- * cpu<id> <user> <nice> <system> <idle> <iowait>
+ * cpu<id> <user> <nice> <system> <idle> <iowait> <irq> <softirq>
  */
 int get_stat(cpuinfo_t *cpu)
 {
 	char *p1, *p2, searchfor[10];
 	int err;
 	
-	if ((err = read_file("/proc/stat", cpu->fd, 0)) < 0) {
-		perror("Error reading /proc/stat");
+	if ((err = read_file("/proc/stat", cpu->fd, 0)) != 0) {
 		return err;
 	}
 
@@ -205,6 +227,9 @@ int get_stat(cpuinfo_t *cpu)
 	cpu->reading->mynice = strtol(p2, &p2, 10);
 	cpu->reading->system = strtol(p2, &p2, 10);
 	cpu->reading->idle = strtol(p2, &p2, 10);
+	cpu->reading->iowait = strtol(p2, &p2, 10);
+	cpu->reading->irq = strtol(p2, &p2, 10);
+	cpu->reading->softirq = strtol(p2, &p2, 10);
 
 	return 0;
 }
@@ -220,22 +245,21 @@ int change_speed(cpuinfo_t *cpu, enum modes mode)
 	
 	if (mode == RAISE) {
 		if ((func == AGGRESSIVE) || (func == LEAPS)) {
-			cpu->current_speed = cpu->max_speed;
+			cpu->speed_index = 0;
 		} else {
-			cpu->current_speed += step;
-			if (cpu->current_speed > cpu->max_speed)
-				cpu->current_speed = cpu->max_speed;
+			if (cpu->speed_index != 0) cpu->speed_index--;
 		} 
 	} else {
 		if ((func == PASSIVE) || (func == LEAPS)) {
-			cpu->current_speed = cpu->min_speed;
+			cpu->speed_index = (cpu->table_size-1);
 		} else {
-			cpu->current_speed -= step;
-			if (cpu->current_speed < cpu->min_speed)
-				cpu->current_speed = cpu->min_speed;
+			if (cpu->speed_index != (cpu->table_size-1))
+				cpu->speed_index++;
 		}
 	}
 
+	cpu->current_speed = cpu->freq_table[cpu->speed_index];
+
 	pprintf(3,"Setting speed to %d\n", cpu->current_speed);
 
 	change_speed_count++;
@@ -275,7 +299,7 @@ int change_speed(cpuinfo_t *cpu, enum modes mode)
 /*
  * The heart of the program... decide to raise or lower the speed.
  */
-int inline decide_speed(cpuinfo_t *cpu)
+enum modes inline decide_speed(cpuinfo_t *cpu)
 {
 	int err;
 	float pct;
@@ -289,15 +313,22 @@ int inline decide_speed(cpuinfo_t *cpu)
 	total = (cpu->reading->user - cpu->last_reading->user) +
 		(cpu->reading->system - cpu->last_reading->system) +
 		(cpu->reading->mynice - cpu->last_reading->mynice) +
-		(cpu->reading->idle - cpu->last_reading->idle);
+		(cpu->reading->idle - cpu->last_reading->idle) +
+		(cpu->reading->iowait - cpu->last_reading->iowait) +
+		(cpu->reading->irq - cpu->last_reading->irq) +
+		(cpu->reading->softirq - cpu->last_reading->softirq);
 
 	if (ignore_nice) { 
 		usage = (cpu->reading->user - cpu->last_reading->user) +
-			(cpu->reading->system - cpu->last_reading->system);
+			(cpu->reading->system - cpu->last_reading->system) +
+			(cpu->reading->irq - cpu->last_reading->irq) +
+			(cpu->reading->softirq - cpu->last_reading->softirq);
 	} else {
 		usage = (cpu->reading->user - cpu->last_reading->user) +
 			(cpu->reading->mynice - cpu->last_reading->mynice) +
-			(cpu->reading->system - cpu->last_reading->system);
+			(cpu->reading->system - cpu->last_reading->system) +
+			(cpu->reading->irq - cpu->last_reading->irq) +
+			(cpu->reading->softirq - cpu->last_reading->softirq);
 	}
 	
 	pct = ((float)usage)/((float)total);
@@ -307,13 +338,28 @@ int inline decide_speed(cpuinfo_t *cpu)
 	if ((pct > ((float)highwater/100.0)) && 
 			(cpu->current_speed != cpu->max_speed)) {
 		/* raise speed to next level */
-		change_speed(cpu, RAISE);
+		return RAISE;
 	} else if ((pct < ((float)lowwater/100.0)) && 
 			(cpu->current_speed != cpu->min_speed)) {
 		/* lower speed */
-		change_speed(cpu, LOWER);
+		return LOWER;
 	}
 	
+	return SAME;
+}
+
+/* 
+ * Abuse glibc's qsort.  Compare function to sort list of frequencies in 
+ * ascending order.
+ */
+int faked_compare(const void *a, const void *b)
+{
+	unsigned long *a1 = (unsigned long *)a;
+	unsigned long *b1 = (unsigned long *)b;
+
+	if (*a1 < *b1) return 1;
+	if (*a1 > *b1) return -1;
+
 	return 0;
 }
 
@@ -322,8 +368,9 @@ int inline decide_speed(cpuinfo_t *cpu)
  */
 int get_per_cpu_info(cpuinfo_t *cpu, int cpuid)
 {
-	char cpustr[100], scratch[100], tmp[10];
+	char cpustr[100], scratch[100], tmp[10], *p1;
 	int fd, err;
+	unsigned long temp;
 	
 	cpu->cpuid = cpuid;
 	cpu->sysfs_dir = (char *)malloc(50*sizeof(char));
@@ -339,28 +386,101 @@ int get_per_cpu_info(cpuinfo_t *cpu, int cpuid)
 	
 	strncpy(scratch, cpu->sysfs_dir, 50);
 	strncat(scratch, "cpuinfo_max_freq", 18);
-	if ((err = read_file(scratch, 0, 1)) < 0) {
-		perror("couldn't open max freq file");
+	if ((err = read_file(scratch, 0, 1)) != 0) {
 		return err;
 	}
 	
 	cpu->max_speed = strtol(buf, NULL, 10);
-
+	
 	strncpy(scratch, cpu->sysfs_dir, 50);
 	strncat(scratch, "cpuinfo_min_freq", 18);
 
-	if ((err = read_file(scratch, 0, 1)) < 0) {
-		perror("couldn't open min freq file");
+	if ((err = read_file(scratch, 0, 1)) != 0) {
 		return err;
 	}
 
 	cpu->min_speed = strtol(buf, NULL, 10);
+
+	/* 
+	 * More error handling, make sure step is not larger than the 
+	 * difference between max and min speeds. If so, truncate it.
+	 */
+	if (step > (cpu->max_speed - cpu->min_speed)) {
+		step = cpu->max_speed - cpu->min_speed;
+	}
+	
+	/* XXXjc read the real current speed */
 	cpu->current_speed = cpu->max_speed;
+	cpu->speed_index = 0;
 
 	strncpy(scratch, cpu->sysfs_dir, 50);
-	strncat(scratch, SYSFS_GOVERNOR, 20);
+	strncat(scratch, "scaling_available_frequencies", 50);
 
-	if ((err = read_file(scratch, 0, 1)) < 0) {
+	if (((err = read_file(scratch, 0, 1)) != 0) || (step_specified)) {
+		/* 
+		 * We don't have scaling_available_frequencies. build the
+		 * table from the min, max, and step values.  the driver
+		 * could ignore these, but we'll represent it this way since
+		 * we don't have any other info.
+		 */
+		cpu->table_size = ((cpu->max_speed-cpu->min_speed)/step) + 1;
+		cpu->table_size += ((cpu->max_speed-cpu->min_speed)%step)?1:0;
+		
+		cpu->freq_table = (unsigned long *)
+			malloc(cpu->table_size*sizeof(unsigned long));
+
+		if (cpu->freq_table == (unsigned long *)NULL) {
+			perror("couldn't allocate cpu->freq_table");
+			return ENOMEM;
+		}
+
+		/* populate the table.  Start at the top, and subtract step */
+		for (temp = 0; temp < cpu->table_size; temp++) {
+			cpu->freq_table[temp] = 
+			((cpu->min_speed<(cpu->max_speed-(temp*step))) ? 
+			 (cpu->max_speed-(temp*step)) :
+			 (cpu->min_speed) );
+		}	
+	} else {
+		/* 
+		 * We do have the file, parse it and build the table from
+		 * there.
+		 */ 
+		/* The format of scaling_available_frequencies (SAF) is:
+		 * "number<space>number2<space>...numberN<space>\n", but this
+		 * can change. So we're relying on the fact that strtoul will 
+		 * return 0 if it can't find anything, and that 0 will never 
+		 * be a real value for the available frequency. 
+		 */
+		p1 = buf;
+		
+		temp = strtoul(p1, &p1, 10);
+		while((temp > 0) && (cpu->table_size < 100)) {
+			cpu->table_size++;
+			temp = strtoul(p1, &p1, 10);
+		}
+	
+		cpu->freq_table = (unsigned long *)
+			malloc(cpu->table_size*sizeof(unsigned long));
+		if (cpu->freq_table == (unsigned long *)NULL) {
+			perror("Couldn't allocate cpu->freq_table\n");
+			return ENOMEM;
+		}
+	
+		p1 = buf;
+		for (temp = 0; temp < cpu->table_size; temp++) {
+			cpu->freq_table[temp] = strtoul(p1, &p1, 10);
+		}
+	}
+
+	/* now lets sort the table just to be sure */
+	qsort(cpu->freq_table, cpu->table_size, sizeof(unsigned long), 
+			&faked_compare);
+	
+	strncpy(scratch, cpu->sysfs_dir, 50);
+	strncat(scratch, "scaling_governor", 20);
+
+	if ((err = read_file(scratch, 0, 1)) != 0) {
 		perror("couldn't open scaling_governors file");
 		return err;
 	}
@@ -378,7 +498,7 @@ int get_per_cpu_info(cpuinfo_t *cpu, int cpuid)
 			close(fd);
 			return err;
 		}
-		if ((err = read_file(scratch, fd, 0)) < 0) {
+		if ((err = read_file(scratch, fd, 0)) != 0) {
 			perror("Error reading back governor file");
 			close(fd);
 			return err;
@@ -402,6 +522,9 @@ int get_per_cpu_info(cpuinfo_t *cpu, int cpuid)
 	 * then 10MHz. Therefore, is the number for max_speed is less than
 	 * 10000, assume the driver is reporting speeds in MHz, not KHz,
 	 * and adjust accordingly.
+	 *
+	 * XXXjc the longhaul driver has been fixed (2.6.5ish timeframe)
+	 * so this should't be needed anymore.  Remove for 1.0?
 	 */
 	cpu->in_mhz = 0;
 	if (cpu->max_speed <= 10000) {
@@ -470,6 +593,68 @@ const char *str_func(void)
 	}
 }
 
+/* Generic x86 cpuid function lifted from kernel sources */
+/*
+ * Generic CPUID function
+ */
+static inline void cpuid(int op, int *eax, int *ebx, int *ecx, int *edx)
+{
+        __asm__("cpuid"
+                : "=a" (*eax),
+                  "=b" (*ebx),
+                  "=c" (*ecx),
+                  "=d" (*edx)
+                : "0" (op));
+}
+
+/* 
+ * a little bit of black magic to try and detect the number of cores per
+ * processor.  This will have to be added on to for every architecture, 
+ * as we learn how to detect them from userspace.  Note: This method
+ * assumes uniform processor/thread ID's.
+ *
+ * you can always override this by using the -c command line option to 
+ * specify the number of threads per core.  If you do so, it will do a static
+ * mapping, uniform for all real processors in the system.  Actually, so 
+ * will this one, because there's no way for me to bind to a processor.
+ * (yet. :)
+ */
+int determine_threads_per_core(int ncpus)
+{
+#ifdef __i386__
+	/* 
+	 * XXXjc fix eventually to run on each processor so you
+	 * can support mixed multi and single-core cpus. Need to know
+	 * how to force ourselves to run on one particular processor.
+	 */
+	int eax,ebx,ecx,edx, num;
+
+	cpuid(1,&eax,&ebx,&ecx,&edx);
+
+	/* 
+	 * Do we support hyperthreading?
+	 * AMD's dual-core will masquerade as HT, so this should work 
+	 * for them too.
+	 */
+	if(edx & 0x08000000) { 
+		/* 
+		 * if so, is it enabled? If so, how many threads 
+		 * are enabled?  Thank you, sandpile.org and the LKML. 
+		 */
+		num = (ebx & 0x00FF0000) >> 16;
+	}
+	/* 
+	 * if num = 0, default to 1.  Other non-multiples of cpus will
+	 * be taken care of later
+	 * XXXjc, rewrite to use sysfs (affected_cpus).
+	 */
+	return ((num)?num:1);
+#endif
+	/* always default to one thread per core */
+	return 1;
+}
+
+
 /* 
  * Main program loop.. parse arguments, sanity chacks, setup signal handlers
  * and then enter main loop
@@ -477,13 +662,14 @@ const char *str_func(void)
 int main(int argc, char **argv)
 {
 	cpuinfo_t *cpu;
-	int ncpus, i, err;
+	int ncpus, i, j, err, num_real_cpus, threads_per_core;
+	enum modes change, change2;
 
 	/* Parse command line args */
 	while(1) {
 		int c;
 
-		c = getopt(argc, argv, "dnvqm:s:p:u:l:h");
+		c = getopt(argc, argv, "dnvqm:s:p:c:u:l:U:L:h");
 		if (c == -1)
 			break;
 
@@ -501,6 +687,15 @@ int main(int argc, char **argv)
  			case 'q':
  				verbosity = -1;
  				break;
+			case 'c':
+				t_per_core = strtol(optarg, NULL, 10);
+				if (t_per_core < 1) {
+					printf("invalid number of cores/proc");
+					help();
+					exit(ENOTSUP);
+				}
+				cores_specified = 1;
+				break;
 			case 'm':
 				func = strtol(optarg, NULL, 10);
 				if ((func < 0) || (func > 3)) {
@@ -517,6 +712,7 @@ int main(int argc, char **argv)
 					help();
 					exit(ENOTSUP);
 				}
+				step_specified = 1;
 				pprintf(2,"Using %dHz step.\n", step);
 				break;
 			case 'p':
@@ -535,7 +731,7 @@ int main(int argc, char **argv)
 					help();
 					exit(ENOTSUP);
 				}
-				pprintf(2,"Using upper limit of %d%%\n",highwater);
+				pprintf(2,"Using upper pct of %d%%\n",highwater);
 				break;
 			case 'l':
 				lowwater = strtol(optarg, NULL, 10);
@@ -544,7 +740,7 @@ int main(int argc, char **argv)
 					help();
 					exit(ENOTSUP);
 				}
-				pprintf(2,"Using lower limit of %d%%\n",lowwater);
+				pprintf(2,"Using lower pct of %d%%\n",lowwater);
 				break;
 			case 'h':
 			default:
@@ -553,9 +749,9 @@ int main(int argc, char **argv)
 		}
 	}
 
-	/* one last thing to check... */
+	/* last things to check... */
 	if (lowwater > highwater) {
-		printf("Invalid: lower limit higher then upper limit!\n");
+		printf("Invalid: lower pct higher than upper pct!\n");
 		help();
 		exit(ENOTSUP);
 	}
@@ -567,7 +763,7 @@ int main(int argc, char **argv)
 		openlog("powernowd", LOG_AUTHPRIV|LOG_PERROR, LOG_DAEMON);
 	
 	/* My ego's pretty big... */
-	pprintf(0,"PowerNow Daemon v%s, (c) 2003-2004 John Clemens\n", 
+	pprintf(0,"PowerNow Daemon v%s, (c) 2003-2005 John Clemens\n", 
 			VERSION);
 
 	/* are we root?? */
@@ -585,8 +781,7 @@ int main(int argc, char **argv)
 	pprintf(1,"  poll interval: %4d ms\n", poll);
 	
 	ncpus = sysconf(_SC_NPROCESSORS_CONF);
-	pprintf(0,"Found %d cpu%s:\n", ncpus, (ncpus==1)?"":"s");
-
+	
 	/* Malloc, initialise data structs */
 	cpu = malloc(sizeof(cpuinfo_t)*ncpus);
 	
@@ -597,14 +792,53 @@ int main(int argc, char **argv)
 	memset(cpu, 0, (sizeof(cpuinfo_t)*ncpus));
 	cpuinfo = &cpu;
 
+	if (cores_specified) {
+		if (ncpus < t_per_core) {
+			printf("\nWARNING: bogus # of thread per core, assuming 1\n");
+			threads_per_core = 1;
+		} else {
+			threads_per_core = t_per_core;
+		}
+	} else { 
+		threads_per_core = determine_threads_per_core(ncpus);
+	}
+	
+	/* We don't support mixed configs yet */
+	if (!ncpus || !threads_per_core || ncpus % threads_per_core) {	
+		printf("WARN: ncpus(%d) is not a multiple of threads_per_core(%d)!\n",
+			ncpus, threads_per_core);
+		printf("WARN: Assuming 1.\n");
+		threads_per_core = 1;
+		//help();
+		//exit(ENOTSUP);
+	}
+	
+	num_real_cpus = ncpus/threads_per_core;
+	for (i=0;i<ncpus;i++) {
+		cpu[i].threads_per_core = threads_per_core;
+	}
+	
+	pprintf(0,"Found %d cpu%s:  -- %d thread%s (or core%s) per physical cpu\n",
+			ncpus/threads_per_core,
+			(ncpus>1)?"s":"",
+			threads_per_core,	
+			(threads_per_core>1)?"s":"", 
+			(threads_per_core>1)?"s":"");
+	
 	for (i=0;i<ncpus;i++) {
 		if ((err = get_per_cpu_info(&cpu[i], i)) != 0) {
-			perror("Couldn't get per-cpu data");
+			printf("\n");
 			goto out;
 		}
-		pprintf(0,"  cpu%d: %dMhz - %dMhz\n", cpu->cpuid, 
-				cpu->min_speed / 1000, 
-				cpu->max_speed / 1000);
+		pprintf(0,"  cpu%d: %dMhz - %dMhz (%d steps)\n", 
+				cpu[i].cpuid,
+				cpu[i].min_speed / 1000, 
+				cpu[i].max_speed / 1000, 
+				cpu[i].table_size);
+		for(j=0;j<cpu[i].table_size; j++) {
+			pprintf(1, "     step%d : %ldMhz\n", j+1, 
+					cpu[i].freq_table[j] / 1000);
+		}
 	}
 	
 	/* now that everything's all set up, lets set up a exit handler */
@@ -616,17 +850,25 @@ int main(int argc, char **argv)
 
 	start_time = time(NULL);
 
+	/* Now the main program loop */
 	while(1) {
 		usleep(poll*1000);
-		for(i=0; i<ncpus; i++) {
-			decide_speed(&cpu[i]);
+		for(i=0; i<num_real_cpus; i+=j) {
+			change = LOWER;
+			/* handle SMT/CMP here */
+			for (j=0; j<cpu[i].threads_per_core; j++) {
+				change2 = decide_speed(&cpu[i+j]);
+				if (change2 > change)
+					change = change2;
+			}
+			if (change != SAME) change_speed(&cpu[i], change);
 		}
 	}
 
 out:
 	printf("PowerNowd encountered and error and could not start.\n");
 	printf("Please make sure that:\n");
-	printf(" - You are running a v2.5/v2.6 kernel or later\n");
+	printf(" - You are running a v2.6.7 kernel or later\n");
 	printf(" - That you have sysfs mounted /sys\n");
 	printf(" - That you have the core cpufreq and cpufreq-userspace\n");
 	printf("   modules loaded into your kernel\n");
